/* ZUES Assembler source file -------------------------------*/

/* ######################################### 
   #                                         
   # (C) JPG 1993	hash.c                  
   #                                         
   # Version 1.4			     
   #                                         
   # Updated 98/11/19 at 19:27:50                                 
   #                                        
   # ident @(#)hash.c	1.4 		     		     
   # 
   # This program is free software; you can redistribute it and/or modify
   # it under the terms of the GNU General Public License as published by
   # the Free Software Foundation; either version 2 of the License, or
   # (at your option) any later version.
   #
   # This program is distributed in the hope that it will be useful,
   # but WITHOUT ANY WARRANTY; without even the implied warranty of
   # MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   # GNU General Public License for more details.
   #
   # You should have received a copy of the GNU General Public License
   # along with this program; if not, write to the Free Software
   # Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
   #                                        
   ######################################### */

/* include header files here		> */

#include "hash.h"

/* source starts here			> */

/* strdup: 	makes a copy of a string including allocating 
		the memory return the string or NULL if the
		allocation fails */
	
char *strdup( char *s )	/* make a duplicate of s */
{
	char *pointer;
	
	/* allocate the memory by calling malloc */
	
	pointer = ( char * ) malloc ( strlen( s ) + 1 );  /* +1 for '\0' */
	if ( pointer != NULL )
		strcpy( pointer, s );

	return pointer;
}

/* hash: form hash value for string s */

unsigned hash( char *s )
{
	unsigned hashval;

	for ( hashval = 0; *s != '\0'; s++ )
		hashval = *s + 31 * hashval;
	
	/* return the integer division of the hashval */

	return hashval % HASHSIZE;
}

/* init_hash:	function to initialise the hash table to point to 
		NULL address spaces
*/

int init_hash( void )
{
	extern struct nlist *hashtab[];

	int return_result = VALID;

	int index;

	for ( index = 0; index < HASHSIZE; index++ ){
		hashtab[ index ] = NULL;
	}

	return ( return_result );

}

/* lookup: loop for s in hashtab */

struct nlist *lookup( char *s )
{
	struct nlist *np;

	for ( np = hashtab[ hash( s ) ]; np != NULL; np = np->next )
		if ( strcmp( s, np->name ) == 0 )
			return np;

	return NULL;
}

/* install: put ( name, defn ) in hashtab */

struct nlist *install( char *name, int at_line )
{
	struct nlist *np;
	unsigned hashval;

	if ( ( np = lookup( name ) ) == NULL ) {    /* not found */ 
		
		/* allocate the memory for the size of the structure */

		np = ( struct nlist * ) malloc( sizeof( *np ) );

		if ( np == NULL || ( np -> name = strdup( name ) ) == NULL )

			/* if the pointer returned is NULL or we cant
			duplicate the name given into the new structure */
			return NULL;

		/* create index from hashing name */

		hashval = hash( name );

		/* reset the symbol reference list */

		np -> reloc_list = NULL;

		/* set the new structures next pointer to point to the
		current pointer in hashval */ 

		np -> next = hashtab[ hashval ];

		/* then set the hashval index to the address of the the
		new structure */

		hashtab[ hashval ] = np;


	} else {	/* its already there */

		if ( assem_state == FIRST_PASS ) {

			char temp_line[ MAXLINE ] = "redefined label: ";
		
			strcat( temp_line, name );
			on_error( temp_line, at_line, WARN );
		}

	}	
	
	return np;
}

/* add_label :	used for ading a defintion to the hashing table 
		its a bit easier to use then install and it has
		added error messages */

int add_label( char label[], int value, int flag, int memflag, int linenum )
{
	extern int diag_state;
	extern int assem_state;

	int return_result = VALID;
	struct nlist *returned_pointer;

	if ( diag_state == FULL )
		printf( "**add_label starts:**\n" );

	if ( ( returned_pointer = install( label, linenum ) ) == NULL ) {
		char temp_line[ MAXLINE ] = "couldn't define label: ";
		
		(void) strcat( temp_line, label );
		on_error( temp_line, linenum, LABEL );
		return_result = INVALID;

	} else {
		if ( diag_state == FULL )
			printf("installing label: %s : val : %d \n", label, value ); 

		if ( ( returned_pointer->defn != value ) && ( memflag == MEM ) && ( assem_state == SECOND_PASS ) ) {

			/* this should already have been declared if its different from the first, 
			   the assembler will not produce the correct code ! */ 

			on_error( "memory locations have changed on final pass", linenum, FINAL ); 

		} else { 			
			returned_pointer->defn = value;
			returned_pointer->flag = flag;
			returned_pointer->memflag = memflag;
			return_result = VALID;
		}
	}

	if ( diag_state == FULL )
		printf( "**add_label ends:**\n\n" );

	return return_result;
}

/* resolve: 	given a name it will return the value associated with it
		or an error if its not found in the final pass */	

int resolve( char *label, int *value)
{
	extern FILE *dump_file;
	extern int diag_state;
	extern int mem_expr;
	extern int address;

	extern int addr_resolve;
	extern int addr_resolve_line;
	extern struct addr_name_list *addr_names;

	extern int linenum;

	int return_result = INVALID;
	struct nlist *point;

	if ( diag_state == FULL )
		printf( "**resolve starts:**\n" );

	if ( ( point = lookup( label ) ) != NULL ) {

		*value=point->defn;
		return_result = point->flag;
		if ( point->memflag == MEM ) {
			mem_expr = VALID;
		}

		if ( assem_state == SECOND_PASS ) {

			/* modified 1/6/94 to produce a reloctaion table for Position Independant Code ( PIC ) */

			fprintf( dump_file, "%s - %d\n", label, address );
			addr_resolve = VALID;
			addr_resolve_line = linenum;
			addr_names = add_name_list( addr_names, label );
		}

	} else {
		if ( assem_state == SECOND_PASS ) {

			on_error( "undefined label in source", linenum, LABEL );
			return_result = INVALID;
		}
	}
	if ( diag_state == FULL )
		printf( "**resolve ends:**\n\n" );
	 	
	return return_result;

}

/* dump_table:	a function to dump the symbol table if required */
		
int dump_table( int header, FILE *output )
{
	struct nlist *pointer;
	char name_str[] = "name";
	char value_str[] = "value";
	char name_und[] = "~~~~";
	char value_und[] = "~~~~~";
	int index;
	int MAX_CHAR = 20, MAX_DIGI = 15;
	if ( header == VALID ) {
		fprintf( output, "\nSymbol Table Dump\n");
		fprintf( output, "~~~~~~~~~~~~~~~~~\n");	
		fprintf( output, "%-*.*s : %*s (VALID?) (MEM LABEL?)\n", MAX_CHAR, MAX_CHAR, name_str, MAX_DIGI, value_str);
		fprintf( output, "%-*.*s : %*s ~~~~~~~~ ~~~~~~~~~~~~\n", MAX_CHAR, MAX_CHAR, name_und, MAX_DIGI, value_und);
	}

	for ( index = 0; index <= HASHSIZE; index++ ) {
		for ( pointer = hashtab[ index ]; pointer != NULL; pointer = pointer->next ) {
			fprintf( output, "%-*.*s : %*d", MAX_CHAR, MAX_CHAR, pointer->name, MAX_DIGI, pointer->defn );
			if ( pointer->flag == VALID ){
				fprintf( output, " VALID   " );
			} else {
				fprintf( output, " INVALID " );
			}
			if ( pointer->memflag == MEM ){
				fprintf( output, " MEM\n");
			} else {
				fprintf( output, " OTHER\n");
			}
		}
	}

}  

/* addr_add:	function that manipulates the symbols address list so
		that relocateable code can be produced by the assembler,
		it should be passed the LAST element in the address list stucture
*/

struct addr_list *addr_add( struct addr_list *previous, int value )
{
	struct addr_list *tmp1;

	struct addr_list *return_result;

	tmp1 = (struct addr_list *) memalign( sizeof ( struct addr_list ), sizeof( struct addr_list ) );
	tmp1 -> address = value;

	tmp1 -> next	= NULL;
	tmp1 -> last	= NULL;

	if ( previous != NULL ) {

		previous -> next = tmp1;

	}

	tmp1 -> last 	 = previous;

	return_result = tmp1;

	return ( return_result );

}

/* add_name_list:	function that will add a name to a linked list
			this function was added to product PIC code
*/

struct addr_name_list *add_name_list( struct addr_name_list *list, char *label )
{
	struct addr_name_list *return_result;

	return_result = (struct addr_name_list *) memalign( sizeof ( struct addr_name_list ), sizeof ( struct addr_name_list ) );

	return_result -> next = NULL;
	return_result -> last = NULL;

	strcpy( return_result -> name, label );	

	if ( list != NULL ) {

		list -> next = return_result;
	
	}

	return_result -> last = list;

	return ( return_result );

}

/* clear_name_list:	function to free the memory used by the name list
			should be used to delete the resolver labels used 
			to generate the relocation table
*/

struct addr_name_list *clear_name_list( struct addr_name_list *list )
{

	struct addr_name_list *tmp, *deletion;

	tmp = list;

	while ( tmp != NULL ) {

		deletion = tmp;

		tmp = tmp -> last;

		(void) free( deletion );

	}

	tmp = NULL;

	return ( tmp );

}	

/* calc_addr_names:	function to calculate the various addresses 
			for the opcodes for the relocation table...
			added for PIC code production
*/

int calc_addr_names( struct addr_name_list *list, int address, int op1, int op2, int offset1, int offset2, int len )
{
	extern int linenum;
	extern int max_address;

	struct addr_name_list *tmp;
	struct nlist *sym;

	int tmp_addr = -1;
	int return_result = VALID;

	/* Get to the start of the list of names */

	if ( list != NULL ) {
		for ( tmp = list; tmp -> last != NULL; tmp = tmp -> last )
			;
	}

	while ( tmp != NULL ) {

		sym = lookup( tmp -> name );

		switch ( op1 ) {
			case e_ :
			case i_e_ :
			case E_ :
			{
				tmp_addr = address + offset1;

				sym -> reloc_list = addr_add( sym -> reloc_list, tmp_addr );

				break;
			}

			default :
			{
				break;
			}
		}

		switch ( op2 ) {
			case e_ :
			case i_e_ :
			case E_ :
			{
				tmp_addr = address + offset2;

				sym -> reloc_list = addr_add( sym -> reloc_list, tmp_addr );

				break;
			}

			default :
			{
				break;
			}

		}

		tmp = tmp -> next;
	}

	return ( return_result );

}

/* reloc_table_dump:	function to dump the symbol relocation table 
			so as to facilitate the production of PIC
			code for the SAM series of machines
*/

int reloc_table_dump( FILE *output )
{
	extern struct addr_list *sorted_list;

	struct nlist *pointer;
	struct addr_list *tmp;

	
	int index, tmp2 = INVALID;
	int MAX_CHAR = 20, MAX_DIGI = 15;

	int return_result = VALID;
	
	for ( index = 0; index <= HASHSIZE; index++ ) {

		for ( pointer = hashtab[ index ]; pointer != NULL; pointer = pointer->next ) {

			if ( pointer -> memflag == MEM ) {

				struct addr_list *index;

				tmp = pointer -> reloc_list;

				/* rewind the addr_list ( contains addresses that reference
				   this variable...
				*/

				if ( tmp != NULL ) {

					for ( ; tmp -> last != NULL; tmp = tmp -> last )
						;

					for ( ; tmp != NULL; tmp = tmp -> next ) {
					
						sorted_list = sort_addr_list( sorted_list, tmp -> address );

					}

				}	

			}

		}

	}

	while ( tmp2 == INVALID ) {

		tmp2 = sort_list( sorted_list );

	}	

	for ( tmp = sorted_list; tmp != NULL; tmp = tmp -> next ) {
					
				
		fprintf( output, "%d\n", tmp -> address );

	}

	return ( return_result );

}

/* sort_addr_list:	function to produce a sorted numerical list of addresses for
			the production of PIC code...
*/

struct addr_list *sort_addr_list( struct addr_list *list, int address )
{
	struct addr_list *return_result = NULL;

	struct addr_list *tmp, *index;

	tmp = (struct addr_list *) memalign( sizeof ( struct addr_list ), sizeof ( struct addr_list ) );
	tmp -> next = NULL;
	tmp -> last = NULL;
	tmp -> address = address;

	if ( list == NULL ) {

		/* i.e. weve been passed a null list ! */

		return_result = tmp;

	} else {

		/* we've been passed a proper list... have to not do some sorting */
		/* Note:: the list should start at the HEAD... */

		return_result = list;

		for ( index = list; index -> next != NULL; index = index -> next ) 
			;

		index -> next = tmp;		

	}

	return ( return_result );

}

/* sort_list:	function to bubble sort an address list...
*/

int sort_list( struct addr_list *list )
{
	struct addr_list *index;

	int tmp, return_result = VALID;

	if ( list != NULL ) {

		for ( index = list; index -> next != NULL; index = index -> next ) {

			if ( ( index -> address ) > ( index -> next -> address ) ) {

				tmp = index -> address;

				index -> address = index -> next -> address;

				index -> next -> address = tmp;

				return_result = INVALID;

			}

		}

	}

	return ( return_result );

}	
