/* ZUES Assembler source header file -------------------------------*/

/* ######################################### 
   #                                         
   # (C) JPG 1993	lexin.l                  
   #                                         
   # Version 1.8			     
   #                                         
   # Updated 98/11/19 at 19:27:51                                 
   #                                        
   # ident @(#)lexin.l	1.8 		     		     
   #                                        
   ######################################### */

/* source starts here			> */

%{
#include "lexin.h"
#undef yywrap

/* yywrap:	function called by lex when EOF is reached, redefined
		to create a second pass through the file */

int yywrap( void ){

	extern int obj_state;
	extern int num_passes;
	extern int no_invalid_expr;
	extern int max_passes;
	extern int offset;
	extern int prev_address;
	extern int address;
	extern int diag_state;
	extern int list_state;
	extern int srec_state;
	extern int max_address;
	extern int zero_len_expr;

	extern FILE *list_file;
	extern FILE *dump_file;
	extern FILE *obj_file;
	extern FILE *rel_file;

	int return_result = 1;

	if ( ( assem_state == FIRST_PASS ) && ( obj_state != NO_OUTPUT ) ) {	

		if ( diag_state == FULL ) 
			printf("\nend of %d pass\n", num_passes);
	
		/* move the file pointer back to the beginning of the file */

		( void ) fseek ( yyin, 0L, SEEK_SET );

		/* if there are no invalid expressions or the number of
		initial resolving passes is more than the maximun number
		of passes as set by the user or default the start the
		second code generation pass */

		if ( ( ( zero_len_expr == VALID ) || ( num_passes >= max_passes ) )  && ( obj_state == FULL ) && ( ( no_invalid_expr == VALID ) || ( num_passes >= max_passes ) ) ) {

			/* set the state to the code generation phase */
			
			if ( diag_state == FULL )  
				printf("Start of final pass\n");

			assem_state = SECOND_PASS;
			linenum = 1;
			max_address = address;
			address = 0;
			prev_address = 0;
			offset = 0;
			return_result = 0;
			(void) clear_words( linenum );
			if ( list_state != NO_OUTPUT ) {
				dump_table( VALID, dump_file );
			}

		} else { 

			/* simply reset the linenumbers and increase 
			the number of passes so far by one */

			linenum = 1;
			address = 0;
			prev_address = 0;
			offset = 0;
			if ( ( no_invalid_expr == INVALID ) && ( diag_state == FULL ) )
				printf( "**invalid expressions in pass:\n");
			no_invalid_expr = VALID;
			num_passes++;
			return_result = 0;
			zero_len_expr = VALID;
			(void) clear_words( linenum );
			
		}
			
	} else { 
		
		/* finish off any srecords we may be doing then 
		   dump the hash table to the screen 
		   and quit */

		if ( srec_state == FULL ) {

			srec_gen( 0, 0, STOP, linenum );

		}

		/* dump the symbol table to a file called <file>.zdmp */
	
		if ( obj_state == FULL ) {

			(void) reloc_table_dump( rel_file );

		}

		if ( list_state == FULL ) {

			if ( diag_state != NO_OUTPUT ) {

				dump_table( INVALID, list_file );

			}
		}
		return_result = 1;
	}
	return return_result;
}
%}
%%
a'	{
	push_assem( yytext, linenum );
	yytext[ 1 ] = '\0';
	push_text( yytext, linenum );
	return(SREGISTERB);
	}
adc	{
	push_text( yytext, linenum );
	push_assem( yytext, linenum );
	return(INSTRUCTION);
	}
add	{
	push_text( yytext, linenum );
	push_assem( yytext, linenum );
	return(INSTRUCTION);
	}
af'	{
	push_text( yytext, linenum );
	push_assem( yytext, linenum );
	return(REGISTERB);
	}
af	{
	push_text( yytext, linenum );
	push_assem( yytext, linenum );
	return(REGISTER);
	}
and	{
	push_op( 'a', linenum );
	push_text( yytext, linenum );
	push_assem( yytext, linenum );
	return(AND);
	}
b'	{
	push_assem( yytext, linenum );
	yytext[ 1 ] = '\0';
	push_text( yytext, linenum );
	return(SREGISTERB);
	}
bc	{
	push_text( yytext, linenum );
	push_assem( yytext, linenum );
	return(REGISTER);
	}
bit	{
	push_text( yytext, linenum );
	push_assem( yytext, linenum );
	return(INSTRUCTION); 
	}
c'	{
	push_assem( yytext, linenum );
	yytext[ 1 ] = '\0';
	push_text( yytext, linenum );
	return(SREGISTERB);
	}
call	{
	push_text( yytext, linenum );
	push_assem( yytext, linenum );
	return(TINSTRUCTION);
	}
ccf	{
	push_text( yytext, linenum );
	push_assem( yytext, linenum );
	return(INSTRUCTION);
	}
cp	{ 
	push_text( yytext, linenum );
	push_assem( yytext, linenum );
	return(INSTRUCTION);
	}
cpd	{
	push_text( yytext, linenum );
	push_assem( yytext, linenum );
	return(INSTRUCTION);
	}
cpdr	{
	push_text( yytext, linenum );
	push_assem( yytext, linenum );
	return(INSTRUCTION);
	}
cpi	{
	push_text( yytext, linenum );
	push_assem( yytext, linenum );
	return(INSTRUCTION);
	}
cpir	{
	push_text( yytext, linenum );
	push_assem( yytext, linenum );
	return(INSTRUCTION);
	}
cpl	{
	push_text( yytext, linenum );
	push_assem( yytext, linenum );
	return(INSTRUCTION);
	}
d'	{
	push_assem( yytext, linenum );
	yytext[ 1 ] = '\0';
	push_text( yytext, linenum );
	return(SREGISTERB);
	}
daa	{
	push_text( yytext, linenum );
	push_assem( yytext, linenum );
	return(INSTRUCTION);
	}
de	{ 
	push_text( yytext, linenum ); 
	push_assem( yytext, linenum );
	return(REGISTER); 
	} 
dec	{
	push_text( yytext, linenum );
	push_assem( yytext, linenum );
	return(INSTRUCTION); 
	}
defb	{
	push_text( yytext, linenum );
	push_assem( yytext, linenum );
	return(DEFB);
	}
defm	{
	push_text( yytext, linenum );
	push_assem( yytext, linenum );
	return(DEFM);
	}
defs	{
	push_text( yytext, linenum );
	push_assem( yytext, linenum );
	return(DEFS);
	}
defw	{
	push_text( yytext, linenum );
	push_assem( yytext, linenum );
	return(DEFW);
	}
deft	{
	push_text( yytext, linenum );
	push_assem( yytext, linenum );
	return(DEFT);
	}
di	{
	push_text( yytext, linenum );
	push_assem( yytext, linenum );
	return(INSTRUCTION);
	}
disp	{
	push_text( yytext, linenum );
	push_assem( yytext, linenum );
	return(DIRECTIVE);
	}
djnz	{
	push_text( yytext, linenum );
	push_assem( yytext, linenum );
	return(INSTRUCTION);
	}
e'	{
	push_assem( yytext, linenum );
	yytext[ 1 ] = '\0';
	push_text( yytext, linenum );
	return(SREGISTERB);
	}
ei	{
	push_text( yytext, linenum );
	push_assem( yytext, linenum );
	return(INSTRUCTION);
	}
ent	{
	push_text( yytext, linenum );
	push_assem( yytext, linenum );
	return(ENT);
	}
end	{
	push_text( yytext, linenum );
	push_assem( yytext, linenum );
	return(END);
	}
equ	{
	push_text( yytext, linenum );
	push_assem( yytext, linenum );
	return(EQU);
	}
ex	{
	push_text( yytext, linenum );
	push_assem( yytext, linenum );
	return(INSTRUCTION);
	}
exx	{
	push_text( yytext, linenum );
	push_assem( yytext, linenum );
	return(INSTRUCTION);
	}
h'	{
	push_assem( yytext, linenum );
	yytext[ 1 ] = '\0';
	push_text( yytext, linenum );
	return(SREGISTERB);
	}
halt	{
	push_text( yytext, linenum );
	push_assem( yytext, linenum );
	return(INSTRUCTION);
	}
hl	{
	push_text( yytext, linenum );
	push_assem( yytext, linenum );
	return(REGISTER);
	}
im	{
	push_text( yytext, linenum );
	push_assem( yytext, linenum );
	return(INSTRUCTION);
	}
in	{
	push_text( yytext, linenum );
	push_assem( yytext, linenum );
	return(INSTRUCTION);
	}
inc	{ 
	push_text( yytext, linenum );
	push_assem( yytext, linenum );
	return(INSTRUCTION);
	}
ind	{
	push_text( yytext, linenum );
	push_assem( yytext, linenum );
	return(INSTRUCTION);
	} 
indr	{
	push_text( yytext, linenum );
	push_assem( yytext, linenum );
	return(INSTRUCTION);
	} 
ini	{
	push_text( yytext, linenum );
	push_assem( yytext, linenum );
	return(INSTRUCTION);
	} 
inir	{
	push_text( yytext, linenum );
	push_assem( yytext, linenum );
	return(INSTRUCTION);
	} 
ix	{ 
	push_text( yytext, linenum );
	push_assem( yytext, linenum );
	return(IREGISTER);
	}
iy	{ 
	push_text( yytext, linenum );
	push_assem( yytext, linenum );
	return(IREGISTER);
	}
jp	{
	push_text( yytext, linenum );
	push_assem( yytext, linenum );
	return(TINSTRUCTION);
	} 
jr	{
	push_text( yytext, linenum );
	push_assem( yytext, linenum );
	return(TINSTRUCTION);
	} 
l'	{
	push_assem( yytext, linenum );
	yytext[ 1 ] = '\0';
	push_text( yytext, linenum );
	return(SREGISTERB);
	} 
ld	{
	push_text( yytext, linenum );
	push_assem( yytext, linenum );
	return(INSTRUCTION);
	} 
ldd	{
	push_text( yytext, linenum );
	push_assem( yytext, linenum );
	return(INSTRUCTION);
	} 
lddr	{
	push_text( yytext, linenum );
	push_assem( yytext, linenum );
	return(INSTRUCTION);
	} 
ldi	{
	push_text( yytext, linenum );
	push_assem( yytext, linenum );
	return(INSTRUCTION);
	} 
ldir	{
	push_text( yytext, linenum );
	push_assem( yytext, linenum );
	return(INSTRUCTION);
	} 
m	{
	push_text( yytext, linenum );
	push_assem( yytext, linenum );
	return(FLAG);
	} 
nc	{
	push_text( yytext, linenum );
	push_assem( yytext, linenum );
	return(FLAG);
	} 
neg	{
	push_text( yytext, linenum );
	push_assem( yytext, linenum );
	return(INSTRUCTION);
	} 
nop	{
	push_text( yytext, linenum );
	push_assem( yytext, linenum );
	return(INSTRUCTION);
	} 
not	{
	push_op( 'n', linenum );
	push_assem( yytext, linenum );
	return(NOT);
	}
nv	{
	push_text( yytext, linenum );
	push_assem( yytext, linenum );
	return(FLAG);
	} 
nz	{
	push_text( yytext, linenum );
	push_assem( yytext, linenum );
	return(FLAG);
	} 
or	{
	push_op( 'o', linenum );
	push_text( yytext, linenum );
	push_assem( yytext, linenum );
	return(OR);
	} 
org	{
	push_text( yytext, linenum );
	push_assem( yytext, linenum );
	return(ORG); 
	} 
otdr	{
	push_text( yytext, linenum );
	push_assem( yytext, linenum );
	return(INSTRUCTION);
	} 
otir	{
	push_text( yytext, linenum );
	push_assem( yytext, linenum );
	return(INSTRUCTION);
	} 
out	{
	push_text( yytext, linenum );
	push_assem( yytext, linenum );
	return(INSTRUCTION);
	} 
outd	{
	push_text( yytext, linenum );
	push_assem( yytext, linenum );
	return(INSTRUCTION);
	} 
outi	{
	push_text( yytext, linenum );
	push_assem( yytext, linenum );
	return(INSTRUCTION);
	} 
p	{
	push_text( yytext, linenum );
	push_assem( yytext, linenum );
	return(FLAG);
	} 
pe	{
	push_text( yytext, linenum );
	push_assem( yytext, linenum );
	return(FLAG);
	} 
po	{
	push_text( yytext, linenum );
	push_assem( yytext, linenum );
	return(FLAG);
	} 
pop	{
	push_text( yytext, linenum );
	push_assem( yytext, linenum );
	return(INSTRUCTION);
	} 
push	{
	push_text( yytext, linenum );
	push_assem( yytext, linenum );
	return(INSTRUCTION);
	} 
r	{
	push_text( yytext, linenum );
	push_assem( yytext, linenum );
	return(SREGISTER);
	} 
res	{ 
	push_text( yytext, linenum );
	push_assem( yytext, linenum );
	return(INSTRUCTION);
	}
ret	{ 
	push_text( yytext, linenum );
	push_assem( yytext, linenum );
	return(TINSTRUCTION);
	}
reti	{
	push_text( yytext, linenum );
	push_assem( yytext, linenum );
	return(INSTRUCTION);
	} 
retn	{
	push_text( yytext, linenum );
	push_assem( yytext, linenum );
	return(INSTRUCTION);
	} 
rl	{
	push_text( yytext, linenum );
	push_assem( yytext, linenum );
	return(INSTRUCTION);
	} 
rla	{
	push_text( yytext, linenum );
	push_assem( yytext, linenum );
	return(INSTRUCTION);
	} 
rlc	{ 
	push_text( yytext, linenum );
	push_assem( yytext, linenum );
	return(INSTRUCTION);
	}
rlca	{
	push_text( yytext, linenum );
	push_assem( yytext, linenum );
	return(INSTRUCTION);
	} 
rld	{
	push_text( yytext, linenum );
	push_assem( yytext, linenum );
	return(INSTRUCTION);
	} 
rr	{
	push_text( yytext, linenum );
	push_assem( yytext, linenum );
	return(INSTRUCTION);
	} 
rra	{
	push_text( yytext, linenum );
	push_assem( yytext, linenum );
	return(INSTRUCTION);
	} 
rrc	{
	push_text( yytext, linenum );
	push_assem( yytext, linenum );
	return(INSTRUCTION);
	} 
rrca	{
	push_text( yytext, linenum );
	push_assem( yytext, linenum );
	return(INSTRUCTION);
	} 
rrd	{
	push_text( yytext, linenum );
	push_assem( yytext, linenum );
	return(INSTRUCTION);
	} 
rst	{
	push_text( yytext, linenum );
	push_assem( yytext, linenum );
	return(INSTRUCTION);
	} 
sbc	{
	push_text( yytext, linenum );
	push_assem( yytext, linenum );
	return(INSTRUCTION);
	} 
scf	{
	push_text( yytext, linenum );
	push_assem( yytext, linenum );
	return(INSTRUCTION);
	} 
set	{
	push_text( yytext, linenum );
	push_assem( yytext, linenum );
	return(INSTRUCTION);
	} 
sla	{
	push_text( yytext, linenum );
	push_assem( yytext, linenum );
	return(INSTRUCTION);
	} 
sp	{
	push_text( yytext, linenum );
	push_assem( yytext, linenum );
	return(REGISTER);
	} 
sra	{
	push_text( yytext, linenum );
	push_assem( yytext, linenum );
	return(INSTRUCTION);
	} 
srl	{
	push_text( yytext, linenum );
	push_assem( yytext, linenum );
	return(INSTRUCTION);
	} 
sub	{
	push_text( yytext, linenum );
	push_assem( yytext, linenum );
	return(INSTRUCTION);
	} 
v	{
	push_text( yytext, linenum );
	push_assem( yytext, linenum );
	return(FLAG);
	} 
xor	{
	push_op( 'x', linenum );
	push_text( yytext, linenum );
	push_assem( yytext, linenum );
	return(XOR);
	} 
z	{
	push_text( yytext, linenum );
	push_assem( yytext, linenum );
	return(FLAG);
	} 
a	{ 
	push_text ( yytext, linenum );
	push_assem( yytext, linenum );
	return(SREGISTER);
	}
b	{ 
	push_text ( yytext, linenum );
	push_assem( yytext, linenum );
	return(SREGISTER);
	}
c	{ 
	push_text ( yytext, linenum );
	push_assem( yytext, linenum );
	return(CEE);
	}
d	{ 
	push_text ( yytext, linenum );
	push_assem( yytext, linenum );
	return(SREGISTER);
	}
e	{ 
	push_text ( yytext, linenum );
	push_assem( yytext, linenum );
	return(SREGISTER); 
	}
h	{ 
	push_text ( yytext, linenum );
	push_assem( yytext, linenum );
	return(SREGISTER);
	}
l	{ 
	push_text ( yytext, linenum );
	push_assem( yytext, linenum );
	return(SREGISTER);
	}
i	{
	push_text( yytext, linenum );
	push_assem( yytext, linenum );
	return(SREGISTER);
	} 
[#]	{
	push_text( yytext, linenum );
	push_assem( yytext, linenum );
	return(HASH);
	}
[$]	{ 
	extern int mem_expr;
	push_assem( yytext, linenum );
	push(address, linenum);
	mem_expr = VALID;	
	return(CURRENT);
	}
[0-9][0-9]*	{ 
		int s;
		s = atoi( yytext );
		push ( s, linenum );
	push_assem( yytext, linenum );
		return(ADDRESS); 
		}
[0-9a-f][0-9a-f]*[h]		{ 
				int temp_expr;
	push_assem( yytext, linenum );
				yytext[yyleng]='\0';
				temp_expr = ( int ) strtol ( yytext, ( char** ) NULL, 16 );
				push(temp_expr, linenum);
				return(HEX); 
				}
[0-8][0-8]*[o]			{ 
				int temp_expr;
	push_assem( yytext, linenum );
				yytext[yyleng]='\0';
				temp_expr = ( int ) strtol ( yytext, ( char** ) NULL, 8 );
				push(temp_expr, linenum);
				return(OCT); 
				}
[0-1][0-1]*[b]			{ 
				int temp_expr;
	push_assem( yytext, linenum );
				yytext[yyleng]='\0';
				temp_expr = ( int ) strtol ( yytext, ( char** ) NULL, 2 );
				push(temp_expr, linenum);
				return(BIN); 
				}
[_?a-zA-Z][_0-9a-zA-Z]*[ ]*[:]	{ 
				char *temp;
				push_assem( yytext, linenum );
				yytext[yyleng-1]='\0';
				strcpy( &last_label, &yytext);
				return(ITOKEN);
				}
[_?a-zA-Z][_0-9a-zA-Z]*   	{
				extern int valid_expr;
				extern int no_invalid_expr;
				extern int assem_state;
				extern int linenum;
				extern int diag_state;
				char *temp;
				int temp_expr, return_result;
				
				push_assem( yytext, linenum );
				yytext[yyleng]='\0';
				if ( ( ( return_result = resolve( &yytext, &temp_expr ) ) != INVALID ) ) {
					if ( diag_state == FULL ) {
						printf("**lex: label: %s - %d\n", yytext, temp_expr );		
					}
					push(temp_expr, linenum);
				} else {
					if( assem_state == FIRST_PASS ){

						valid_expr = INVALID;
						no_invalid_expr = INVALID;

					} else {

						char temp_line[ MAXLINE ] = "cant resolve label: ";
				
						(void) strcat( temp_line, yytext );
						on_error( temp_line, linenum, LABEL );
					}
				}
		
	 			return( TOKEN );
				}
["]	{ 
	int lscount = 0, lscount2 = 0;
	int lschar;
	char lsmore;
	char lstring[ MAXSTRING ], temp[ MAXSTRING ];

	temp[ lscount2++ ] = '"';
	
	while ( ( lschar=input( ) ) != '"' ) {
		temp[ lscount2 ] = lschar;
		if ( ( lstring[ lscount ] = lschar ) == '\\' ) {
			lsmore = input( );
			temp[ ++lscount2 ] = lsmore;
			switch ( lsmore ) {
				case 'n': {
					lstring [ lscount ] = '\n';
					}
					break;
				case 't': {
					lstring [ lscount ] = '\t';
					}
					break;
				case 'f': {
					lstring [ lscount ] = '\f';
					}
					break;
				case 'b': {
					lstring [ lscount ] = '\b';
					}
					break;
				case 'r': {
					lstring [ lscount ] = '\r';
					}
					break;
				default :
					lstring [ ++lscount ] = lsmore;
					break;
			}
		}
		lscount++;
		lscount2++;
	}
	lstring[ lscount ] = '\0';
	temp[ lscount2++ ] = '"';
	temp[ lscount2 ] = '\0';
	push_str( lstring, linenum);
	push_assem( temp, linenum );
	return(STRING); 
	}
[']	{ 
	extern int linenum;
	int lscount = 0, lscount2 = 0;
	int lschar;
	char lstring[ MAXSTRING ], temp[ MAXSTRING ];

	temp[ lscount2++ ] = '\''; 
	
	while ( ( lschar=input() ) != '\''){
		lstring[ lscount++ ] = lschar;
		temp[ lscount2++ ] = lschar;
	}
	if ( lscount >= 3 ) 
		on_error( "cannot determine \\ character", linenum, EXPR ); 
	if ( lstring[ 0 ]=='\\' && ( lscount <= 2 ) ) {
		switch ( lstring[ 1 ] ) {
			case 'n': {
				push( '\n', linenum );
				}
				break;
			case 'r': {
				push( '\r', linenum );
				}
				break;
			case 't': {
				push( '\t', linenum );
				}
				break;
			case 'f': {
				push( '\f', linenum );
				}
				break;
			case 'b': {
				push( '\b', linenum );
				}
				break;
			default :
				push( lstring[ 1 ], linenum );
				break;
		}
	} else {
		push( lstring[ 0 ], linenum );
	}
	temp[ lscount2++ ] = '\'';
	temp[ lscount2 ] = '\0';
	push_assem( temp, linenum );
	return(ADDRESS); 
	}
[,]	{
	push_assem( yytext, linenum );
	return(COMMA);
	} 
[\n]	{
	return(NEWLINE);
	} 
[;]	{ 
	extern int not_empty;

	int count = 0;
	char c, temp_line[ MAXLINE ];

	push_assem( yytext, linenum );

	while ( ( c = input() ) != '\n'){
		temp_line[ count++ ] = c;
	}
	temp_line[ count ] = '\0';
	push_assem( temp_line, linenum );
	not_empty = VALID;
		
	return(NEWLINE); 
	}
[(]	{
	push_assem( yytext, linenum );
	push_op ( yytext[0], linenum );
	return(OPENBR);
	} 
[)]	{
	push_assem( yytext, linenum );
	push_op ( yytext[0], linenum );
	return(CLOSEBR);
	} 
[%]	{
	push_assem( yytext, linenum );
	push_op( yytext[0], linenum );
	return(MOD);
	}
[+]	{ 
	push_assem( yytext, linenum );
	push_op ( yytext[0], linenum );
	return(PLUS);
	}
[-]	{
	push_assem( yytext, linenum );
	push_op ( yytext[0], linenum );
	return(MINUS);
	}
[*]	{
	push_assem( yytext, linenum );
	push_op ( yytext[0], linenum );
	return(MULT);
	}
[/]	{
	push_assem( yytext, linenum );
	push_op ( yytext[0], linenum );
	return(DIVIDE);
	}
[\[]	{
	push_assem( yytext, linenum );
	push_op ( yytext[0], linenum );
	return(OPSQR);
	}
[\]]	{
	push_assem( yytext, linenum );
	push_op ( yytext[0], linenum );
	return(CLSQR);
	}
[\t]    { 
	push_assem( yytext, linenum ); 
	}
[ ][ ]* {
		extern int format_space;
		char space[] = " \0";
		if ( format_space == VALID ) {
			push_assem( space, linenum );
		} else {	
			push_assem( yytext, linenum );
		}
	}
