/* ZUES Assembler source file -------------------------------*/

/* ######################################### 
   #                                         
   # (C) JPG 1993	stack.c                  
   #                                         
   # Version 1.3			     
   #                                         
   # Updated 98/11/19 at 19:27:51                                 
   #                                        
   # ident @(#)stack.c	1.3 		     		     
   # 
   # This program is free software; you can redistribute it and/or modify
   # it under the terms of the GNU General Public License as published by
   # the Free Software Foundation; either version 2 of the License, or
   # (at your option) any later version.
   #
   # This program is distributed in the hope that it will be useful,
   # but WITHOUT ANY WARRANTY; without even the implied warranty of
   # MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   # GNU General Public License for more details.
   #
   # You should have received a copy of the GNU General Public License
   # along with this program; if not, write to the Free Software
   # Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
   #                                        
   ######################################### */

/* include header files here		> */

#include "stack.h"

/* source starts here			> */

/* print_op:	diagnostic function to dump the operator stack */

int print_op( linenum )
{
	extern int diag_state;
	int index, return_result;
	
	if ( diag_state == FULL ) {
		printf( "**print_op starts:**\n");
		for ( index = 0; index < sp_op; index++ ) {
			printf ( "stack op val: %c at: %d\n", val_op[ index ], index );
		}
		printf( "**print_op ends:**\n\n");
	}
	return VALID;
}

/* print_stack:	diagnostic function to dump the number stack */

int print_stack( linenum )
{
	extern int diag_state;
	int index, return_result;

	if ( diag_state == FULL ) {
		printf( "**print_stack starts:**\n");
		for ( index = 0; index < sp_num; index++ ) {
			printf ( "stack val: %d at: %d\n", val_num[ index ], index );
		}
		printf( "**print_stack ends:**\n\n");
	}
	return VALID;
}

/* push: push number onto stack */

int push( int number, int line_num )
{
	int return_result;

	if ( sp_num < MAXVAL) {
		val_num[sp_num++] = number;
		return_result = VALID;
	
	} else {			/* the stacks full */

		on_error( "push: stack full cant push number", line_num, SYSTEM );
		return_result = INVALID;
	}
	
	return return_result;
		
}

/* pop: pop and return top value of the number stack 
	does not give an error for stack empty because
	this should be detected by the calling function */

int pop( int *number )
{
	int return_result;

	if ( sp_num > 0 ) { 		/* the stacks not empty */
	
		*number = val_num[--sp_num];
		return_result = VALID;

	} else {				/* the stack is empty */
		*number = 0;		/* make sure garbage isnt returned */
		return_result = INVALID;
	}

	return return_result;

}

/* save_stack:	a function to save the entire contents of the number
		stack, used by functions which wish to alter the stack
		without actually destroying it, pushes it in reverse
		order so that the recovery function is simpler. */

int save_stack( int linenum )
{
	int return_result;
	int temp_number;
	int temp_sp;

	while ( pop( &temp_number ) != INVALID ) {

		if ( sp_num_save < MAXVAL) {
			val_num_save[ sp_num_save++ ] = temp_number;
			return_result = VALID;
		
		} else {			/* the stacks full */
	
			on_error( "save_stack: stack full cant save number", linenum, SYSTEM );
			return_result = INVALID;
		}
	}

	/* reset the stack pointer */
	sp_num = 0;
	temp_sp = sp_num_save;
	if ( return_result == VALID ) {
		int index;
		for ( index = --temp_sp; index >= 0; index-- ) {
			if ( push( val_num_save[ index ], linenum ) != INVALID ) {
				return_result = VALID;
			} else {
				on_error( "recover_stack: stack full cant push number", linenum, SYSTEM );
			}
		}
	}
	

	return return_result;
}

/* recover_stack:	a function to recover the stack from the saved
			array, will return invalid if the stack becomes 
			full */

int recover_stack( int linenum )
{
	int return_result;
	int index, temp_number;

	/* reset the stack pointer */
	sp_num = 0;

	for ( index = --sp_num_save ; index >= 0; index-- ) {
		if ( push( val_num_save[ index ], linenum ) != INVALID ) {
			return_result = VALID;
		} else {
			on_error( "recover_stack: stack full cant push number", linenum, SYSTEM );
		}
	}

	sp_num_save = 0;

	return return_result;
}


/* ------------------------------------------------------------ */

/* push_text: push a text string onto the text stack */

int push_text( char string[], int line_num )
{
	int return_result;

	if ( sp_text < MAXVAL) { /* the stacks not full */

		strcpy( val_text[sp_text++], string ) ;

		return_result = VALID;
	} else {			/* the stack is full */

		on_error( "push_text: stack full cant push string", line_num, SYSTEM );

		return_result = INVALID;

	}

	return return_result;

}


/* pop_text:	pop and return top value of the text stack 
		returns invalid if the stacks empty */

int pop_text( char string[] )
{
	int return_result;

	if ( sp_text > 0 ) {	/* the stacks not empty */

		strcpy( string, val_text[--sp_text] ) ;
		
		return_result = VALID;

	} else {			/* the stacks empty */

		return_result = INVALID;
	}

	return return_result;

}

/* ------------------------------------------------------------ */

/* push_assem: push a string onto the assembler stack */

int push_assem( char string[], int line_num )
{
	int return_result;

	if ( sp_assem < MAXVAL) { /* the stacks not full */

		strcpy( val_assem[sp_assem++], string ) ;
		
		return_result = VALID;
	} else {			/* the stack is full */

		on_error( "push_assem: stack full cant push assembler", line_num, SYSTEM );

		return_result = INVALID;

	}

	return return_result;

}


/* pop_assem:	pop and return top value of the assem stack 
		returns invalid if the stacks empty */

int pop_assem( char string[] )
{
	int return_result;

	if ( sp_assem > 0 ) {	/* the stacks not empty */

		strcpy( string, val_assem[--sp_assem] ) ;
		
		return_result = VALID;

	} else {			/* the stacks empty */

		return_result = INVALID;
	}

	return return_result;

}

/* ------------------------------------------------------------ */

/* push_str: push a string onto the string stack */

int push_str( char string[], int line_num )
{
	int return_result;

	if ( sp_str < MAXVAL) { /* the stacks not full */

		strcpy( val_str[sp_str++], string ) ;
		
		return_result = VALID;
	} else {			/* the stack is full */

		on_error( "push_str: stack full cant push string", line_num, SYSTEM );

		return_result = INVALID;

	}

	return return_result;

}


/* pop_str:	pop and return top value of the string stack 
		returns invalid if the stacks empty */

int pop_str( char string[] )
{
	int return_result;

	if ( sp_str > 0 ) {	/* the stacks not empty */

		strcpy( string, val_str[--sp_str] ) ;
		
		return_result = VALID;

	} else {			/* the stacks empty */

		return_result = INVALID;
	}

	return return_result;

}

/* ------------------------------------------------------------ */

/* push_op: push operator onto stack */

int push_op( char operator, int line_num )
{
	extern int diag_state;
	int return_result;

	if ( diag_state == FULL )	
		printf("****push_op: pushed %c **\n", operator  );

	if ( sp_op < MAXVAL) {
		val_op[sp_op++] = operator;
		return_result = VALID;

	} else { 		/* the stacks full */

		on_error( "push_op: stack full cant push operator", line_num, SYSTEM );
		return_result = INVALID;

	}

	return return_result;

}


/* pop_op: 	pop and return top operator of the operator stack 
		does not give an error for stack empty as this
		should be detected by the calling function */
	
int pop_op(char *operator)
{
	int return_result;

	if ( sp_op > 0 ) {	/* the stacks not empty */

		*operator = val_op[--sp_op];
		return_result = VALID;
		
	} else {			/* the stack is empty */

		return_result = INVALID;

	}

	return return_result;
}

/* ------------------------------------------------------------ */

/* clear_stacks:	clears the operator and number stacks to 
			ensure nothing is left on after each line
			returns invalid and the top of the number
			stack if it is not empty */

int clear_stacks( int *temp_number )
{
	int return_result;

	if ( pop( temp_number ) == VALID ) { 

		/* theres something still on the stack */

		return_result = INVALID;

	} else {
		/* theres nothing left on the stack */
		/* temp number => zero due to pop   */

		return_result = VALID;

	}

	sp_num = 0; 		/* set the stack pointers to zero */
	val_num[0]=0;
	sp_op = 0;
	valid_expr = VALID;	/* set the marker back to VALID */
	sp_num_save = 0;

	return return_result;	
	
}

/* clear_words: clears down the text, string & assem stacks
		shouldn't really be needed but its here just
		in case */

int clear_words ( int line_num )
{
	extern int sp_text;
	extern int sp_str;
	extern int sp_assem;

	sp_text = 0;
	sp_str = 0;
	sp_assem = 0;

	return VALID;
}

/* invert_stack:	function to invert the entire number stack
			not used for expressions but for printing
			lists of expressions */

int invert_stack ( int line_num )
{
	int temp_array[MAXVAL+1],count=0,index=0,temp_expr=0;

	while ( ( pop(&temp_expr)) == VALID ) {

		temp_array[count++]=temp_expr;

		}

	for ( index=0 ; index <= (count-1) ; index++)
		push(temp_array[index], line_num );

	return VALID;
}

/* invert_words:	function used to invert a stack of strings
			either the string, assem or text stacks may
			be inverted by specifying the type */
	
	
int invert_words ( int type, int line_num )
{
	int return_result = VALID;
	char temp_array[ MAXVAL + 1 ][ MAXSTRING ], temp_str[ MAXSTRING ];
	
	switch ( type ) {
	case A_STRING : { 
		int count=0,index=0;

		while ( ( pop_str( temp_str ) ) == VALID ) {

			(void) strcpy( temp_array[ count++ ], temp_str );
		}
	
			for ( index=0; index <= ( count-1 ); index++ )
				return_result = push_str( temp_array[ index ], line_num );
		}
		break;

	case A_ASSEM : { 
		int count=0,index=0;
	
		while ( ( pop_assem( temp_str ) ) == VALID ) {

			(void) strcpy( temp_array[ count++ ], temp_str );
		}
	
			for ( index=0; index <=( count-1 ); index++ ){
				return_result = push_assem( temp_array[ index ], line_num );
			}
		}
		break;

	case A_TEXT : { 
		int count=0,index=0;
	
		while ( ( pop_text( temp_str ) ) == VALID ) {

			(void) strcpy( temp_array[ count++ ], temp_str );
		}
	
			for ( index=0; index <= ( count-1 ); index++ )
				return_result = push_text( temp_array[ index ], line_num );
		}
		break;

	default : {
		on_error( "invert_words: invalid option", line_num, SYSTEM );
		}
	}
	return return_result;
}
